/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <string.h>

#include <glib.h>
#include <gnome.h>

#include "acache.h"
#include "article-db.h"
#include "debug.h"
#include "server.h"
#include "folderlist.h"
#include "group.h"
#include "gui.h"
#include "article.h"
#include "articlelist.h"
#include "util.h"

/***
****
***/

PanCallback * article_deleted = NULL;

/***
****
***/

void
article_init (void)
{
	article_deleted = pan_callback_new ();
}

void
article_isolate (Article* article)
{
	GSList *l = NULL;
	Article *parent = article->parent;

	if (parent)
		parent->threads = g_slist_remove (parent->threads, article);

	for (l=article->threads; l!=NULL; l=l->next) {
		Article *child = ARTICLE(l->data);
		child->parent = parent;
		if (parent)
			parent->threads = g_slist_append (parent->threads, child);
	}

	g_slist_free (article->threads);
	article->threads = NULL;
	article->parent = NULL;
}

/**
***
**/

void
article_destructor (PanObject* o)
{
	Article* article = ARTICLE(o);
        g_return_if_fail (article!=NULL);

	if (article->dirty)
	{
		gpointer article_db = ahdb_ref (article->group);
		debug1 (DEBUG_DB, "saving article [%s] to disk", article->subject);
		ahdb_save (article_db, article, FALSE);
		ahdb_unref (article_db);
	}

	/* clean up article parts */
	debug1 (DEBUG_PAN_OBJECT, "article_destructor: %p", article);
	replace_gstr (&article->subject, NULL);
	replace_gstr (&article->author, NULL);
	replace_gstr (&article->message_id, NULL);
	replace_gstr (&article->references, NULL);
	replace_gstr (&article->newsgroups, NULL);
	replace_gstr (&article->reply_to, NULL);
	replace_gstr (&article->followup_to, NULL);
	replace_gstr (&article->filename, NULL);
	g_slist_free (article->threads);

	/* clean up parent */
	pan_object_destructor (o);
}

void
article_constructor (Article* article)
{
	g_return_if_fail (article!=NULL);

	/* init the parent class */
	pan_object_constructor (PAN_OBJECT(article), article_destructor);

	/* init the article bits */
	debug1 (DEBUG_PAN_OBJECT, "article_constructor: %p", article);
	article->self_passes_filter = FALSE;
	article->tree_passes_filter = FALSE;
	article->dirty = FALSE;
	article->unread_children = 0;
	article->group = NULL;
	article->parts = 0;
	article->part = 0;
	article->linecount = 0;
	article->crosspost_qty = 0;
	article->state = 0;
	article->date = (time_t)0;
	article->number = 0;
	article->subject = NULL;
	article->author = NULL;
	article->message_id = NULL;
	article->filename = NULL;
	article->references = NULL;
	article->reply_to = NULL;
	article->followup_to = NULL;
	article->newsgroups = NULL;
	article->parent = NULL;
	article->threads = NULL;
}

Article*
article_new (void)
{
	Article* article = g_new (Article, 1);
	article_constructor (article);
	return article;
}

/***
****
***/

gboolean
article_is_read (const Article * a)
{
	g_return_val_if_fail (a!=NULL, FALSE);
	g_return_val_if_fail (a->group!=NULL, FALSE);

	return newsrc_is_article_read (a->group->newsrc, a->number);
}

/* returns the number of articles marked unread. */
static int
article_set_unread_impl (Article * article)
{
	int unread_qty = 0;
	const gboolean was_read = newsrc_mark_article (article->group->newsrc,
	                                               article->number,
	                                               FALSE);
	if (was_read) {
		articlelist_article_read_changed (article);
		++unread_qty;
	}

	// FIXME: I don't think this special case is such a great idea.
	// Special case: if it's a multipart, mark 1..n as read too
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				unread_qty += article_set_unread_impl (child);
		}
	}

	return unread_qty;
}

/* returns the number of articles marked read. */
static int
article_set_read_impl (Article * article)
{
	int read_qty = 0;
	const gboolean was_read = newsrc_mark_article (article->group->newsrc,
	                                               article->number,
	                                               TRUE);
	if (!was_read) {
		++read_qty;
		articlelist_article_read_changed (article);
	}

	// FIXME: I don't think this special case is such a great idea.
	// Special case: if it's a multipart, mark 1..n as read too
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				read_qty += article_set_read_impl (child);
		}
	}

	return read_qty;
}

void
articles_set_read (Article ** articles, int article_qty, gboolean read)
{
	//FIXME: this fails if the batch of articles have differing groups
	int i;
	int read_inc;

	g_return_if_fail (articles != NULL);

	for (read_inc=i=0; i<article_qty; ++i)
	{
		Article * article = articles[i];

		g_return_if_fail (article != NULL);
		g_return_if_fail (article->group != NULL);

		if (read)
			read_inc += article_set_read_impl (article);
		else 
			read_inc -= article_set_unread_impl (article);
	}

	if (read_inc)
		group_inc_article_read_qty (articles[0]->group, read_inc);
}

void
article_set_read (Article * article, gboolean read)
{
	articles_set_read (&article, 1, read);
}
        
        
/***
****
***/

static void
article_delete_impl (gpointer article_db,
                     Article * article,
                     int * read,
                     int * total)
{
	g_return_if_fail (article != NULL);
	g_return_if_fail (article->group != NULL);

	/* notify listeners before the delete takes place */
	pan_callback_call (article_deleted, article, NULL);

	/* update counts */
	++*total;
	if (article_is_read (article))
		++*read;

	/* isolate */
	article_isolate (article);

	/* erase from cache */
	acache_delete (article->message_id);

	/* erase from db */
	ahdb_erase (article_db, article, NULL);

	/* erase from memory */
	pan_object_unref (PAN_OBJECT(article));
}

void
articles_delete (Article      ** articles,
                 int             article_qty,
                 gboolean        sync)
{
	if (article_qty > 0)
	{
		int total = 0;
		int read = 0;
		int i;
		Group * group; // FIXME: what if articles in different groups?
		gpointer article_db;

		/* sanity checks */
		g_return_if_fail (articles != NULL);

		group = articles[0]->group;

		/* update the database */
		article_db = ahdb_ref (group);
		for (i=0; i!=article_qty; ++i)
			article_delete_impl (article_db, articles[i], &read, &total);
		if (sync)
			ahdb_sync (article_db);
		ahdb_unref (article_db);

		/* update group totals */
		if (total)
			group_inc_article_qty (group, -total);
		if (read)
			group_inc_article_read_qty (group, -read);
	}
}


/**
*** PUBLIC MUTATORS
**/

void
article_set_dirty (Article   * article,
                   gboolean    dirty)
{
	g_return_if_fail (article != NULL);
	article->dirty = dirty;
	if (article->dirty)
		articlelist_update_node (article);
}

void
article_set_references (Article         * article,
	                const gchar     * references)
{
	g_return_if_fail (article != NULL);
	replace_gstr (&article->references, g_strdup(references));
	article_set_dirty (article, TRUE);
}

void
article_set_reply_to (Article         * article,
	              const gchar     * reply_to)
{
	g_return_if_fail (article != NULL);
	replace_gstr (&article->reply_to, g_strdup(reply_to));
	article_set_dirty (article, TRUE);
}

void
article_set_followup_to (Article         * article,
                         const gchar     * followup_to)
{
	g_return_if_fail (article != NULL);
	replace_gstr (&article->followup_to, g_strdup(followup_to));
	article_set_dirty (article, TRUE);
}

void
article_set_newsgroups (Article     * article,
	                const gchar * newsgroups)
{
	g_return_if_fail (article != NULL);
	replace_gstr (&article->newsgroups, g_strdup(newsgroups));
	article_set_dirty (article, TRUE);
}

void
article_add_flag (Article  * article,
                  gushort    flag)
{
	g_return_if_fail (article != NULL);
	article->state |= flag;
	article_set_dirty (article, TRUE);
}


void
article_remove_flag (Article         * article,
                     gushort           flag)
{
	g_return_if_fail (article != NULL);
	article->state &= ~flag;
	article_set_dirty (article, TRUE);
}


/**
***
**/

gchar*
article_get_thread_message_id (const Article* article)
{
	char * pch;

	g_return_val_if_fail (article!=NULL, NULL);

	if (article->references)
	{
		pch = get_next_token_str (article->references, ' ', NULL);
	}
	else /* top of the thread */
	{
		pch = g_strdup (article->message_id);
	}

	return pch;
}

static void
article_get_entire_thread_impl (Article* top, GPtrArray* setme)
{
	GSList *l;
	g_ptr_array_add (setme, top);
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_impl (ARTICLE(l->data), setme);
}

static void
article_get_thread_impl (Article * article,
                         GPtrArray * setme,
                         gboolean children_only)
{
	Article* top = article;

	if (top != NULL)
	{
		if (!children_only)
			while (top->parent != NULL)
				top = top->parent;

		pan_g_ptr_array_reserve (setme, 128);
		article_get_entire_thread_impl (top, setme);
	}
}

void
article_get_subthread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, TRUE);
}

void
article_get_entire_thread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, FALSE);
}

void
articles_unref (Article ** articles, size_t qty)
{
	g_return_if_fail (articles != NULL);

	if (qty != 0)
	{
		size_t i;
		gpointer article_db = NULL;
		Group * const group = articles[0]->group;

		/* Loop through each item, saving the dirty ones.
		   This is faster than having each unreffed article
		   open and close the database on its own. */
		for (i=0; i!=qty; ++i)
		{
			Article * a = ARTICLE(articles[i]);
			if (a!=NULL && a->dirty)
			{
	       			if (article_db == NULL)
					article_db = ahdb_ref (group);

				if (a->group == group)
				{
					ahdb_save (article_db, a, FALSE);
					article_set_dirty (a, FALSE);
				}
			}

			pan_object_unref (PAN_OBJECT(a));
		}

		if (article_db != NULL)
			ahdb_unref (article_db);
	}
}


#if 0
/*
 * 1. open source db
 * 2. copy article from source db to folder db
 * 3. erase article from source db
 *
 */
static void
article_move_to_folder (Article *article,
			Group   *group,
			Group   *folder)
{
	g_return_if_fail (article != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

	article_copy_to_folder (article, group, folder);

	article_delete (group, article, FALSE);
}

void
article_copy_to_folder (Article *article,
			Group   *group,
			Group   *folder)
{
//	gpointer srcdb  = NULL;
//	gpointer destdb = NULL;

	g_return_if_fail (article != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

//	srcdb = ahdb_ref (server, from);
//	destdb = ahdb_ref (folder_server, folder);

	ahdb_save (destdb, article, TRUE);

//	ahdb_unref (srcdb);
//	ahdb_sync (destdb);
	ahdb_unref (destdb);
}


typedef struct
{
	Server *server;
	Group  *group;
	GSList *list_of_articles;
} ArticleMoveStruct;

void
article_copy_list_to_folder (GSList *articles,
			     Group *group)
{

}

static void
article_move_list_destroy_cb (GtkObject *object, gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	GSList *l;

	for (l=ams->list_of_articles; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		g_slist_foreach (l2, article_free_gfunc, NULL);
		g_slist_free (l2);
	}
	g_slist_free (ams->list_of_articles);
	g_free (ams);
}

static void
article_move_list_clicked_cb (GnomeDialog *dialog,
                              int button,
                               gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	Server *server = ams->server;
	Group  *group = ams->group;
	Group *folder = NULL;
	GSList *l;

	g_return_if_fail (server != NULL);
	g_return_if_fail (group != NULL);


	/* Cancel button */
	if (button == 1)
	{
		pan_lock ();
		article_move_list_destroy_cb (GTK_OBJECT (dialog), NULL);
		gtk_widget_destroy (GTK_WIDGET (dialog));
		pan_unlock ();
		return;
	}

	folder = gtk_object_get_data (GTK_OBJECT (dialog), "selectedfolder");

	for (l=ams->list_of_articles; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		GSList *tmp = NULL;

		for (tmp=l2; tmp!=NULL; tmp=tmp->next)
		{
			Article *article = ARTICLE(tmp->data);

			article_move_to_folder (article, server, group, folder);
		}
	}

	pan_lock ();
	gtk_widget_destroy (GTK_WIDGET (dialog));
	pan_unlock ();
}

void
article_move_list_to_folder (GSList *articles,
                             Server *server,
                             Group *group)
{
	GnomeDialog *dialog = folderlist_picker_new ();

	ArticleMoveStruct *ams = g_new (ArticleMoveStruct, 1);

	ams->server = server;
	ams->group = group;
	ams->list_of_articles = articles;

	/* connect signals */
	pan_lock();
	gtk_signal_connect (GTK_OBJECT (dialog), "clicked",
			    GTK_SIGNAL_FUNC (article_move_list_clicked_cb),
			    ams);
//	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
//			    GTK_SIGNAL_FUNC (article_move_list_destroy_cb),
//			    ams);
	pan_unlock();
			    
	gui_popup_draw (GTK_WIDGET (dialog), Pan.window);
}
#endif
